"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * @module parjs/internal
 */
/** */
const combinators_1 = require("./implementation/combinators");
const implementation_1 = require("./implementation");
const functions_1 = require("./implementation/functions");
const reply_1 = require("../reply");
const soft_1 = require("./implementation/combinators/alternatives/soft");
const act_1 = require("./implementation/combinators/map/act");
const index_1 = require("../index");
const isolate_1 = require("./implementation/combinators/special/isolate");
const convertible_literal_1 = require("./convertible-literal");
const isFunction = require("lodash/isFunction");
function wrap(action) {
    return new ParjsParser(action);
}
function flattenNestedArrays(arr) {
    if (!Array.isArray(arr)) {
        return [arr];
    }
    let items = [];
    for (let item of arr) {
        if (Array.isArray(item)) {
            items.push(...flattenNestedArrays(item));
        }
        else {
            items.push(item);
        }
    }
    return items;
}
class ParjsParser extends implementation_1.BaseParjsParser {
    get backtrack() {
        return wrap(new combinators_1.PrsBacktrack(this.action)).withName("backtrack");
    }
    get state() {
        let ret = wrap(new combinators_1.PrsProject(this.action, (r, s) => s));
        return ret.withName("state");
    }
    get q() {
        return wrap(new combinators_1.PrsQuieten(this.action)).withName("quiet");
    }
    get soft() {
        return wrap(new soft_1.PrsSoften(this.action)).withName("soften");
    }
    get not() {
        return wrap(new combinators_1.PrsInverse(this.action)).withName("not");
    }
    get str() {
        return wrap(new combinators_1.PrsStringify(this.action)).withName("str");
    }
    thenChoose(selector) {
        return wrap(new combinators_1.PrsChoose(this.action, selector)).withName("thenChoose");
    }
    between(preceding, proceeding) {
        let bet;
        preceding = convertible_literal_1.ConversionHelper.convert(preceding);
        proceeding = convertible_literal_1.ConversionHelper.convert(proceeding);
        if (proceeding) {
            bet = preceding.q.then(this).then(proceeding.q);
        }
        else {
            bet = preceding.q.then(this).then(preceding.q);
        }
        return bet.withName("between");
    }
    mustCapture(failType = reply_1.ReplyKind.HardFail) {
        return wrap(new combinators_1.PrsMustCapture(this.action, failType)).withName("mustCapture");
    }
    maybe(x) {
        if (arguments.length > 0) {
            return wrap(new combinators_1.PrsMaybe(this.action, x)).withName("maybe");
        }
        else {
            return wrap(new combinators_1.PrsMaybe(this.action, implementation_1.QUIET_RESULT)).withName("maybe");
        }
    }
    or(...others) {
        others = others.map(p => convertible_literal_1.ConversionHelper.convert(p));
        return wrap(new combinators_1.PrsAlternatives([this, ...others].map(x => x.action))).withName("or");
    }
    map(f) {
        //f is (result, userState) => any if this.isLoud
        //f is (userState) => any otherwise
        let mapper;
        if (this.isLoud) {
            mapper = f;
        }
        else {
            mapper = (result, state) => f(state);
        }
        return wrap(new combinators_1.PrsProject(this.action, mapper)).withName("map");
    }
    each(f) {
        //f is (result, userState) => void if this.isLoud
        //f is (userState) => void otherwise.
        let mapper;
        if (this.isLoud) {
            mapper = f;
        }
        else {
            mapper = (result, userState) => f(userState);
        }
        return wrap(new act_1.PrsEach(this.action, mapper)).withName("each");
    }
    then(...args) {
        let next;
        let unpack = false;
        if (args.length === 0) {
            return this;
        }
        else if (Array.isArray(args[0])) {
            next = args[0];
        }
        else if (index_1.Parjs.helper.isParser(args[0]) || typeof args[0] === "string" || args[0] instanceof RegExp) {
            unpack = true;
            next = args;
        }
        next = next.map(p => convertible_literal_1.ConversionHelper.convert(p));
        let actions = [this.action, ...next.map(x => x.action)];
        let seqParse = wrap(new combinators_1.PrsSequence(actions));
        let loudCount = actions.filter(x => x.isLoud).length;
        let ret;
        if (unpack && loudCount === 1) {
            ret = seqParse.map(x => x[0]);
        }
        else if (unpack && loudCount === 0) {
            ret = seqParse.q;
        }
        else {
            ret = seqParse;
        }
        return ret.withName("then");
    }
    many(minSuccesses = 0, maxIters = Infinity) {
        return wrap(new combinators_1.PrsMany(this.action, maxIters, minSuccesses)).withName("many");
    }
    manyTill(till, tillOptional = false) {
        till = convertible_literal_1.ConversionHelper.convert(till);
        if (isFunction(till)) {
            return this.must(till, undefined, reply_1.ReplyKind.SoftFail).many();
        }
        return wrap(new combinators_1.PrsManyTill(this.action, till.action, tillOptional)).withName("manyTill");
    }
    manySepBy(sep, maxIterations = Infinity) {
        sep = convertible_literal_1.ConversionHelper.convert(sep);
        return wrap(new combinators_1.PrsManySepBy(this.action, sep.action, maxIterations)).withName("manySepBy");
    }
    exactly(count) {
        return wrap(new combinators_1.PrsExactly(this.action, count)).withName("exactly");
    }
    result(r) {
        return wrap(new combinators_1.PrsProjectConst(this.action, r)).withName("result");
    }
    cast() {
        return this;
    }
    must(condition, name = "(unnamed condition)", fail = reply_1.ReplyKind.HardFail) {
        let cond = condition;
        if (!this.isLoud) {
            cond = (x, state) => condition(state);
        }
        return wrap(new combinators_1.PrsMust(this.action, cond, fail, name)).withName("must");
    }
    mustNotBeOf(options, fail = reply_1.ReplyKind.HardFail) {
        return this.must(x => options.indexOf(x) < 0, `none of: ${options.join(", ")}`, fail).withName("mustNotBeOf");
    }
    mustBeOf(options, fail = reply_1.ReplyKind.HardFail) {
        return this.must(x => options.indexOf(x) >= 0, `one of: ${options.join(", ")}`, fail).withName("mustBeOf");
    }
    mustBeNonEmpty(fail = reply_1.ReplyKind.HardFail) {
        return this.must(functions_1.Predicates.nonEmpty, `be non-empty`, fail).withName("mustBeNonEmpty");
    }
    withName(newName) {
        this.displayName = newName;
        return this;
    }
    isolateState(x) {
        return wrap(new isolate_1.PrsIsolate(this.action, x)).withName(`isolated: ${this.action.displayName}`);
    }
    flatten() {
        return this.map(x => flattenNestedArrays(x));
    }
    splat() {
        return this.map((arr) => {
            let result = {};
            for (let item of arr) {
                Object.assign(result, item);
            }
            return result;
        });
    }
}
exports.ParjsParser = ParjsParser;

//# sourceMappingURL=instance.js.map