"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * @module parjs/internal
 */
/** */
const instance_1 = require("./instance");
const parsers_1 = require("./implementation/parsers");
const combinators_1 = require("./implementation/combinators");
const action_1 = require("./implementation/action");
const reply_1 = require("../reply");
const int_1 = require("./implementation/parsers/numbers/int");
const float_1 = require("./implementation/parsers/numbers/float");
const basic_trace_visualizer_1 = require("./implementation/basic-trace-visualizer");
const char_indicators_1 = require("./implementation/functions/char-indicators");
const common_1 = require("../common/common");
const convertible_literal_1 = require("./convertible-literal");
const _defaults = require("lodash/defaults");
function wrap(action) {
    return new instance_1.ParjsParser(action);
}
class NoUnicodeError extends Error {
    constructor() {
        super("The optional Unicode character recognizers haven't been loaded. Please load them by importing/requiring 'parjs/unicode'");
    }
}
exports.ConditionalUnicode = new class InfoContainer {
    constructor() {
        this._codeInfo = null;
        this._charInfo = null;
    }
    get CodeInfo() {
        if (!this._codeInfo) {
            throw new NoUnicodeError();
        }
        return this._codeInfo;
    }
    set CodeInfo(info) {
        this._codeInfo = this._codeInfo || info;
    }
    get CharInfo() {
        if (!this._charInfo) {
            throw new NoUnicodeError();
        }
        return this._charInfo;
    }
    set CharInfo(info) {
        this._charInfo = this._charInfo || info;
    }
}();
class ParjsHelper {
    isParser(obj) {
        return obj instanceof instance_1.ParjsParser;
    }
    isParserAction(obj) {
        return obj instanceof action_1.ParjsBasicAction;
    }
}
exports.ParjsHelper = ParjsHelper;
class ParjsParsers {
    constructor() {
        this.helper = new ParjsHelper();
        this.visualizer = basic_trace_visualizer_1.BasicTraceVisualizer();
    }
    get spaces1() {
        return this.space.many(1).str.withName("spaces1");
    }
    get letter() {
        return this.charWhere(char_indicators_1.AsciiCharInfo.isLetter).withName("letter");
    }
    get anyChar() {
        return wrap(new parsers_1.PrsStringLen(1)).withName("anyChar");
    }
    get uniLetter() {
        return this.charWhere(exports.ConditionalUnicode.CharInfo.isUniLetter).withName("uniLetter");
    }
    get digit() {
        return this.charWhere(char_indicators_1.AsciiCharInfo.isDecimal).withName("digit");
    }
    get uniDigit() {
        return this.charWhere(exports.ConditionalUnicode.CharInfo.isUniDecimal).withName("uniDigit");
    }
    get hex() {
        return this.charWhere(char_indicators_1.AsciiCharInfo.isHex).withName("hex");
    }
    get uniLower() {
        return this.charWhere(exports.ConditionalUnicode.CharInfo.isUniLower).withName("uniLower");
    }
    get lower() {
        return this.charWhere(char_indicators_1.AsciiCharInfo.isLower).withName("lower");
    }
    get upper() {
        return this.charWhere(char_indicators_1.AsciiCharInfo.isUpper).withName("upper");
    }
    get uniUpper() {
        return this.charWhere(exports.ConditionalUnicode.CharInfo.isUniUpper).withName("uniUpper");
    }
    get newline() {
        return wrap(new parsers_1.PrsNewline(null)).withName("newline");
    }
    get uniNewline() {
        return wrap(new parsers_1.PrsNewline(exports.ConditionalUnicode.CodeInfo)).withName("uniNewline");
    }
    get space() {
        return this.charWhere(char_indicators_1.AsciiCharInfo.isSpace).withName("space");
    }
    get uniSpace() {
        return this.charWhere(exports.ConditionalUnicode.CharInfo.isUniSpace).withName("uniSpace");
    }
    get whitespaces() {
        return this.charWhere(c => char_indicators_1.AsciiCharInfo.isSpace(c) || char_indicators_1.AsciiCharInfo.isNewline(c)).many().str.withName("whitespaces");
    }
    get uniSpaces() {
        return this.uniSpace.many().str.withName("uniSpaces");
    }
    get nop() {
        return this.result(undefined).q.withName("nop");
    }
    get rest() {
        return wrap(new parsers_1.PrsRest()).withName("rest");
    }
    get eof() {
        return wrap(new parsers_1.PrsEof()).withName("eof");
    }
    get position() {
        return wrap(new parsers_1.PrsPosition()).withName("position");
    }
    get state() {
        return wrap(new parsers_1.PrsState()).withName("state");
    }
    seqObject(...specs) {
        let parsers = [];
        let keys = [];
        for (let spec of specs) {
            for (let key in spec) {
                if (!Object.prototype.hasOwnProperty.call(spec, key))
                    continue;
                parsers.push(spec[key]);
                keys.push(key);
            }
        }
        return this.seq(...parsers).map(arr => {
            let obj = {};
            for (let i = 0; i < keys.length; i++) {
                obj[keys[i]] = arr[i];
            }
            return obj;
        });
    }
    late(resolver) {
        return wrap(new combinators_1.PrsLate(() => resolver().action, true)).withName("late");
    }
    choose(selector) {
        return this.result(undefined).thenChoose((v, st) => selector(st));
    }
    any(...parsers) {
        let parsers2 = parsers.map(x => convertible_literal_1.ConversionHelper.convert(x));
        return wrap(new combinators_1.PrsAlternatives(parsers2.map(x => x.action))).withName("any");
    }
    seq(...parsers) {
        let parsers2 = parsers.map(x => convertible_literal_1.ConversionHelper.convert(x));
        return wrap(new combinators_1.PrsSequence(parsers.map(x => x.action))).withName("seq");
    }
    charWhere(predicate, property) {
        return wrap(new parsers_1.PrsCharWhere(predicate, property)).withName(`charWhere`);
    }
    anyCharOf(options) {
        return this.charWhere(x => common_1.Es6.strIncludes(options, x), `any of ${options}`).withName("anyCharOf");
    }
    noCharOf(options) {
        return this.charWhere(x => !common_1.Es6.strIncludes(options, x)).withName("noCharOf");
    }
    string(str) {
        return wrap(new parsers_1.PrsString(str)).withName("string");
    }
    anyStringOf(...strs) {
        return wrap(new parsers_1.PrsStringOf(strs)).withName("anyStringOf");
    }
    stringLen(length) {
        return wrap(new parsers_1.PrsStringLen(length)).withName("stringLen");
    }
    regexp(regex) {
        let flags = `${common_1.Es6.regexFlags(regex).replace("g", "").replace("y", "")}y`;
        let stickyRegex = new RegExp(regex.source, flags);
        return wrap(new parsers_1.PrsRegexp(stickyRegex)).withName("regexp");
    }
    result(x) {
        return wrap(new parsers_1.PrsResult(x)).withName("result");
    }
    fail(expecting = "", kind = reply_1.ReplyKind.SoftFail) {
        return wrap(new parsers_1.PrsFail(kind, expecting)).withName("fail");
    }
    int(options) {
        options = _defaults({}, options, {
            base: 10,
            allowSign: true
        });
        return wrap(new int_1.PrsInt(options)).withName("int");
    }
    float(options) {
        options = _defaults({}, options, {
            allowImplicitZero: true,
            allowExponent: true,
            allowSign: true,
            allowFloatingPoint: true
        });
        return wrap(new float_1.PrsFloat(options)).withName("float");
    }
}
exports.ParjsParsers = ParjsParsers;

//# sourceMappingURL=static.js.map